from __future__ import annotations

import dataclasses
import datetime
import pathlib
import re
import uuid
from typing import Any, Tuple, Callable
from typing import Dict
from typing import Optional, List, Union
from uuid import uuid4

import httpx
import lxml.html
import pprintpp as pprint
import requests
import starlette.testclient
from lxml.etree import _Element
from pytest import FixtureRequest
from requests import PreparedRequest
from responses import POST
from responses import RequestsMock, CallbackResponse, FalseBool

from spinta import api
from spinta import auth
from spinta import commands
from spinta.auth import create_client_file, get_clients_path, yaml, yml
from spinta.client import add_client_credentials
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.testing.config import create_config_path
from spinta.testing.context import TestContext
from spinta.testing.context import create_test_context
from spinta.testing.datasets import Sqlite


def create_test_client(
    rc_or_context: Union[RawConfig, TestContext],
    request: FixtureRequest = None,
    *,
    config: Dict[str, Any] = None,
    scope: List[str] = None,
    raise_server_exceptions: bool = True,
) -> TestClient:
    if isinstance(rc_or_context, RawConfig):
        rc = rc_or_context
        if config:
            rc = rc.fork(config)
        context = create_test_context(rc, request, name='pytest/client')
    else:
        if config is not None:
            raise NotImplementedError()
        context = rc_or_context
    if context.loaded:
        if config is not None:
            raise RuntimeError("Context already loaded, can't override config.")
    else:
        context.load()
    app = api.init(context)
    client = TestClient(
        context,
        app,
        base_url='https://testserver',
        raise_server_exceptions=raise_server_exceptions,
    )
    if scope:
        client.authorize(scope)
    return client


def create_remote_server(
    rc: RawConfig,
    tmp_path: pathlib.Path,
    responses: RequestsMock,
    *,
    url: str = 'https://example.com/',
    client: str = 'client',
    secret: str = 'secret',
    scopes: List[str] = None,
    credsfile: Union[bool, pathlib.Path] = None,
) -> RemoteServer:

    def remote(request: PreparedRequest):
        path = request.url[len(url.rstrip('/')):]
        resp = app.request(
            request.method,
            path,
            headers=dict(request.headers),
            content=request.body,
        )
        return resp.status_code, resp.headers, resp.content

    confdir = create_config_path(tmp_path / 'config')
    rc = rc.fork({
        'config_path': confdir,
        'default_auth_client': None,
    })
    context = create_test_context(rc)
    app = create_test_client(context)

    if scopes:
        client_file, client = create_client_file(
            get_clients_path(confdir),
            client_id=str(uuid.uuid4()),
            name=client,
            secret=secret,
            scopes=scopes,
            add_secret=True,
        )
        secret = client['client_secret']
        client = client['client_name']

    if credsfile:
        if credsfile is True:
            credsfile = tmp_path / 'credentials.cfg'
        add_client_credentials(
            credsfile, url,
            client=client,
            secret=secret,
            scopes=scopes,
        )

    responses._registry.add(
        CustomCallbackResponse(
            url=re.compile(re.escape(url) + r'.*'),
            method=POST,
            callback=remote,
            content_type='application/json',
            match_querystring=FalseBool(),
            match=(),
        )
    )

    return RemoteServer(
        app=app,
        url=url,
        client=client,
        secret=secret,
        credsfile=credsfile,
    )


@dataclasses.dataclass
class RemoteServer:
    app: TestClient
    url: str
    client: str = None
    secret: str = None
    credsfile: pathlib.Path = None


class TestClientResponse(httpx.Response):
    template: str
    context: Dict[str, Any]

    # So pytest does not treat this as a test collection (since it starts with Test prefix)
    __test__ = False


class TestClient(starlette.testclient.TestClient):
    context: Union[Context, TestContext]

    def __init__(self, context, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = context
        self._scopes = []

    def authmodel(self, model: str, actions: List[str], creds=None):
        scopes = commands.get_model_scopes(self.context, model, actions)
        self.authorize(scopes, creds=creds)

    def authorize(self, scopes: Optional[list] = None, creds=None, strict_set: bool = False):
        # Calling `authorize` multiple times, will preserve previous scopes if strict_set is False.
        if strict_set:
            self._scopes = scopes if scopes is not None else []
        else:
            self._scopes += [s for s in (scopes or []) if s not in self._scopes]

        if creds:
            # Request access token using /auth/token endpoint.
            resp = self.request('POST', '/auth/token', auth=creds, data={
                'grant_type': 'client_credentials',
                'scope': ' '.join(self._scopes),
            })
            assert resp.status_code == 200, resp.text
            token = resp.json()['access_token']
        else:
            # Create access token using private key.
            context = self.context
            private_key = auth.load_key(context, auth.KeyType.private)
            client = 'test-client'
            expires_in = int(datetime.timedelta(days=10).total_seconds())
            token = auth.create_access_token(context, private_key, client, expires_in, scopes=self._scopes)

        self.headers.update({
            'Authorization': f'Bearer {token}'
        })

    def unauthorize(self):
        self._scopes = []
        if 'Authorization' in self.headers:
            del self.headers['Authorization']

    def getdata(self, *args, **kwargs):
        resp = self.get(*args, **kwargs)
        assert resp.status_code == 200, f'status_code: {resp.status_code}, response: {resp.text}'
        resp = resp.json()
        assert '_data' in resp, pprint.pformat(resp)
        return resp['_data']


def get_html_tree(resp: requests.Response) -> Union[
    _Element,
    lxml.html.HtmlMixin,
]:
    return lxml.html.fromstring(resp.text)


def configure_remote_server(
    cli,
    local_rc: RawConfig,
    rc: RawConfig,
    tmp_path: pathlib.Path,
    responses,
    remove_source: bool = True
):
    invoke_props = [
        'copy',
        '--access', 'open',
        '-o', tmp_path / 'remote.csv',
        tmp_path / 'manifest.csv',
    ]
    if remove_source:
        invoke_props.append('--no-source')
    cli.invoke(local_rc, invoke_props)

    # Create remote server with PostgreSQL backend
    remote_rc = rc.fork({
        'manifests': {
            'default': {
                'type': 'tabular',
                'path': str(tmp_path / 'remote.csv'),
                'backend': 'default',
                'mode': local_rc.get('manifests', 'default', 'mode')
            },
        },
        'backends': ['default'],
    })
    return create_remote_server(
        remote_rc,
        tmp_path,
        responses,
        scopes=[
            'spinta_set_meta_fields',
            'spinta_getone',
            'spinta_getall',
            'spinta_search',
            'spinta_insert',
            'spinta_patch',
            'spinta_delete',
            'spinta_changes'
        ],
        credsfile=True,
    )


def create_rc(rc: RawConfig, tmp_path: pathlib.Path, db: Sqlite, mode: str = 'internal') -> RawConfig:
    return rc.fork({
        'manifests': {
            'default': {
                'type': 'tabular',
                'path': str(tmp_path / 'manifest.csv'),
                'backend': 'sql',
                'keymap': 'default',
                'mode': mode
            },
        },
        'backends': {
            'sql': {
                'type': 'sql',
                'dsn': db.dsn,
            },
        },
        # tests/config/clients/3388ea36-4a4f-4821-900a-b574c8829d52.yml
        'default_auth_client': '3388ea36-4a4f-4821-900a-b574c8829d52',
    })


def create_client(rc: RawConfig, tmp_path: pathlib.Path, geodb: Sqlite):
    rc = create_rc(rc, tmp_path, geodb)
    context = create_test_context(rc)
    return create_test_client(context)


class CustomCallbackResponse(CallbackResponse):
    def __init__(
        self,
        method: str,
        url: Union[re.Pattern[str], str],
        callback: Callable[[Any], Any],
        stream: Optional[bool] = None,
        content_type: Optional[str] = "text/plain",
        **kwargs: Any,
    ) -> None:
        super().__init__(method, url, callback, stream, content_type, **kwargs)

    def matches(self, request: "PreparedRequest") -> Tuple[bool, str]:
        self.method = request.method

        if not self._url_matches(self.url, str(request.url)):
            return False, "URL does not match"

        valid, reason = self._req_attr_matches(self.match, request)
        if not valid:
            return False, reason

        return True, ""


def get_yaml_data(path: pathlib.Path) -> dict:
    data = {}
    if path.exists():
        data = yaml.load(path)
    return data


def create_old_client_file(client_path: pathlib.Path, data: dict, file_name: str = None):
    if file_name is None:
        file_name = data.get('client_id', str(uuid4()))

    yml.dump(data, client_path / f'{file_name}.yml')




import base64
from decimal import Decimal
from typing import Any, overload, Optional

from spinta.core.ufuncs import ufunc, Expr
from spinta.datasets.backends.sql.ufuncs.components import FileSelected
from spinta.datasets.backends.sql.ufuncs.result.components import SqlResultBuilder
from spinta.exceptions import UnableToCast
from spinta.types.datatype import Integer, String
from spinta.types.file.components import FileData
from spinta.ufuncs.basequerybuilder.components import Selected


@overload
@ufunc.resolver(SqlResultBuilder)
def cast(env: SqlResultBuilder) -> Any:
    return env.call('cast', env.prop.dtype, env.this)


@overload
@ufunc.resolver(SqlResultBuilder, String, int)
def cast(env: SqlResultBuilder, dtype: String, value: int) -> str:
    return str(value)


@overload
@ufunc.resolver(SqlResultBuilder, String, type(None))
def cast(env: SqlResultBuilder, dtype: String, value: Optional[Any]) -> str:
    return ''


@overload
@ufunc.resolver(SqlResultBuilder, Integer, Decimal)
def cast(env: SqlResultBuilder, dtype: Integer, value: Decimal) -> int:
    return env.call('cast', dtype, float(value))


@overload
@ufunc.resolver(SqlResultBuilder, Integer, float)
def cast(env: SqlResultBuilder, dtype: Integer, value: float) -> int:
    if value % 1 > 0:
        raise UnableToCast(dtype, value=value, type=dtype.name)
    else:
        return int(value)


@ufunc.resolver(SqlResultBuilder, Expr)
def file(env: SqlResultBuilder, expr: Expr) -> FileData:
    """Post query file data processor

    Will be called with _FileSelected kwargs and no args.
    """
    kwargs: FileSelected
    args, kwargs = expr.resolve(env)
    assert len(args) == 0, args
    name = env.data[kwargs['name'].item]
    content = env.data[kwargs['content'].item]
    if isinstance(content, str):
        content = content.encode('utf-8')
    if content is not None:
        content = base64.b64encode(content).decode()
    return {
        '_id': name,
        # TODO: Content probably should not be returned if not explicitly
        #       requested in select list.
        '_content': content,
    }


@ufunc.resolver(SqlResultBuilder, Expr)
def swap(env: SqlResultBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return env.call('swap', *args, *kwargs)


@overload
@ufunc.resolver(SqlResultBuilder, Selected, Selected)
def point(env: SqlResultBuilder, x: Selected, y: Selected) -> Expr:
    x = env.data[x.item]
    y = env.data[y.item]
    return f'POINT ({x} {y})'



import math

from starlette.convertors import Convertor, register_url_convertor


class SpintaFloatConvertor(Convertor[float]):
    regex = r"-?[0-9]+(\.[0-9]+)?"

    def convert(self, value: str) -> float:
        return float(value)

    def to_string(self, value: float) -> str:
        value = float(value)
        assert not math.isnan(value), "NaN values are not supported"
        assert not math.isinf(value), "Infinite values are not supported"
        return ("%0.20f" % value).rstrip("0").rstrip(".")


register_url_convertor('spinta_float', SpintaFloatConvertor())



import io
import json
import pathlib
from typing import Dict, Any, Iterator

import dask
import numpy as np
import requests
from dask.dataframe import DataFrame
from lxml import etree

from spinta import commands
from spinta.components import Context, Property, Model
from spinta.core.ufuncs import Expr
from spinta.datasets.backends.dataframe.components import DaskBackend, Csv, Xml, Json
from spinta.datasets.backends.dataframe.ufuncs.components import TabularResource
from spinta.datasets.backends.dataframe.ufuncs.query.components import DaskDataFrameQueryBuilder
from spinta.datasets.backends.helpers import handle_ref_key_assignment
from spinta.datasets.components import Resource
from spinta.datasets.helpers import get_enum_filters, get_ref_filters
from spinta.datasets.keymaps.components import KeyMap
from spinta.datasets.utils import iterparams
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.dimensions.param.components import ResolvedParams
from spinta.exceptions import PropertyNotFound, NoExternalName, ValueNotInEnum
from spinta.manifests.components import Manifest
from spinta.manifests.dict.helpers import is_list_of_dicts, is_blank_node
from spinta.types.datatype import PrimaryKey, Ref, DataType, Boolean, Number, Integer, DateTime
from spinta.typing import ObjectData
from spinta.ufuncs.basequerybuilder.components import Selected
from spinta.ufuncs.helpers import merge_formulas
from spinta.ufuncs.resultbuilder.components import ResultBuilder
from spinta.utils.data import take
from spinta.utils.schema import NA


def _resolve_expr(context: Context, row: Any, sel: Selected) -> Any:
    if sel.item is None:
        val = None
    else:
        val = row[sel.item]
    env = ResultBuilder(context).init(val, sel.prop, row)
    return env.resolve(sel.prep)


def _aggregate_values(data, target: Property):
    if target is None or target.list is None:
        return data

    key_path = target.place
    key_parts = key_path.split('.')

    # Drop first part, since if nested prop is part of the list will always be first value
    # ex: from DB we get {"notes": [{"note": 0}]}
    # but after fetching the value we only get [{"note": 0}]
    # so if our place is "notes.note", we need to drop "notes" part
    if len(key_parts) > 1:
        key_parts = key_parts[1:]

    def recursive_collect(sub_data, depth=0):
        if depth < len(key_parts):
            if isinstance(sub_data, list):
                collected = []
                for item in sub_data:
                    collected.extend(recursive_collect(item, depth))
                return collected
            elif isinstance(sub_data, dict) and key_parts[depth] in sub_data:
                return recursive_collect(sub_data[key_parts[depth]], depth + 1)
        else:
            return [sub_data]

        return []

    # Start the recursive collection process
    return recursive_collect(data, 0)


def _get_row_value(context: Context, row: Any, sel: Any) -> Any:
    if isinstance(sel, Selected):
        if isinstance(sel.prep, Expr):
            val = _resolve_expr(context, row, sel)
        elif sel.prep is not NA:
            val = _get_row_value(context, row, sel.prep)
        else:
            if sel.item in row.keys():
                val = row[sel.item]
            else:
                raise PropertyNotFound(
                    sel.prop.model,
                    property=sel.prop.name,
                    external=sel.prop.external.name,
                )

        if enum := get_prop_enum(sel.prop):
            if val is None:
                pass
            elif str(val) in enum:
                item = enum[str(val)]
                if item.prepare is not NA:
                    val = item.prepare
            else:
                raise ValueNotInEnum(sel.prop, value=val)

        return val
    if isinstance(sel, tuple):
        return tuple(_get_row_value(context, row, v) for v in sel)
    if isinstance(sel, list):
        return [_get_row_value(context, row, v) for v in sel]
    if isinstance(sel, dict):
        return {k: _get_row_value(context, row, v) for k, v in sel.items()}
    return sel


@commands.load.register(Context, DaskBackend, dict)
def load(context: Context, backend: DaskBackend, config: Dict[str, Any]):
    pass


@commands.prepare.register(Context, DaskBackend, Manifest)
def prepare(context: Context, backend: DaskBackend, manifest: Manifest, **kwargs):
    pass


@commands.bootstrap.register(Context, DaskBackend)
def bootstrap(context: Context, backend: DaskBackend):
    pass


def _select(prop: Property, df: DataFrame):
    if prop.external is None or not prop.external.name:
        raise NoExternalName(prop)
    column = prop.external.name
    if column not in df.columns:
        raise PropertyNotFound(
            prop.model,
            property=prop.name,
            external=prop.external.name,
        )
    return Selected(
        item=column,
        prop=prop,
    )


def _select_primary_key(model: Model, df: DataFrame):
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = _select(prop, df)
    else:
        result = [
            _select(prop, df)
            for prop in pkeys
        ]
    return Selected(prop=model.properties['_id'], prep=result)


def _parse_xml_loop_model_properties(value, added_root_elements: list, model_props: dict, namespaces: dict):
    new_dict = {}

    # Go through each prop source with xpath of root path
    for prop in model_props.values():
        v = value.xpath(prop["source"], namespaces=namespaces)
        new_value = None
        if v:
            v = v[0]
            if isinstance(v, etree._Element):

                # Check if prop has pkeys (means its ref type)
                # If pkeys exists iterate with xpath and get the values
                pkeys = prop["pkeys"]

                if pkeys and prop["source"].endswith(('.', '/')):
                    ref_keys = []
                    for key in pkeys:
                        ref_item = v.xpath(key, namespaces=namespaces)
                        if ref_item:
                            ref_item = ref_item[0]
                            if isinstance(ref_item, etree._Element):
                                if ref_item.text:
                                    ref_item = ref_item.text
                                elif ref_item.attrib:
                                    ref_item = ref_item.attrib
                                else:
                                    ref_item = None
                            elif not ref_item:
                                ref_item = None
                            ref_keys.append(str(ref_item))
                        if len(ref_keys) == 1:
                            new_value = ref_keys[0]
                        else:
                            new_value = ref_keys
                else:
                    if v.text:
                        new_value = str(v.text)
                    elif v.attrib:
                        new_value = v.attrib.values()
                    else:
                        new_value = None
            elif v:
                new_value = str(v)
            else:
                new_value = None
        new_dict[prop["source"]] = new_value

    added_root_elements.append(value)
    return new_dict


def _parse_xml(data, source: str, model_props: Dict, namespaces={}):
    added_root_elements = []
    iterparse = etree.iterparse(data, events=['start'], remove_blank_text=True)
    _, root = next(iterparse)

    if not source.startswith(('/', '.')):
        source = f'/{source}'

    values = root.xpath(source, namespaces=namespaces)
    for value in values:
        if value not in added_root_elements:
            yield _parse_xml_loop_model_properties(value, added_root_elements, model_props, namespaces)


def _parse_json(data, source: str, model_props: Dict):
    data = json.loads(data)
    source_list = source.split(".")
    # Prepare source for blank nodes
    blank_nodes = is_blank_node(data)
    if source == ".":
        source_list = ["."]
    else:
        if blank_nodes:
            source_list = ["."] + source_list
    if blank_nodes:
        for prop in model_props.values():
            if prop["root_source"][0] != '.':
                prop["root_source"].insert(0, '.')

    yield from _parse_json_with_params(data, source_list, model_props)


def _parse_json_with_params(data, source: list, model_props: dict):

    def get_prop_value(items, pkeys: list, prop_source: list):
        new_value = items

        # Remove empty values from split()
        prop_source = [prop for prop in prop_source if prop]
        if pkeys:
            ref_keys = []
            for key in pkeys:
                split = key.split(".")
                split = [prop for prop in split if prop]
                new_value = items
                for id, prop in enumerate(split):
                    if isinstance(new_value, dict):
                        if prop in new_value.keys():
                            new_value = new_value[prop]
                        else:
                            break
                    else:
                        break
                    if id == len(split) - 1:
                        ref_keys.append(new_value)
            if len(ref_keys) == 1:
                return ref_keys[0]
            else:
                return ref_keys
        else:
            for id, prop in enumerate(prop_source):
                if isinstance(new_value, dict):
                    if prop in new_value.keys():
                        new_value = new_value[prop]
                    else:
                        return None
                else:
                    return None
                if id == len(prop_source) - 1:
                    return new_value

    def traverse_json(items, current_value: Dict, current_path: int = 0, in_model: bool = False):
        last_path = source[:current_path + 1]
        c_path = source[current_path]
        if c_path.endswith("[]"):
            c_path = c_path[:-2]
        if isinstance(items, dict):
            if in_model or c_path == ".":
                item = items
            else:
                if c_path in items.keys():
                    item = items[c_path]
                else:
                    return
            if isinstance(item, list) and is_list_of_dicts(item):
                yield from traverse_json(item, current_value, current_path, True)
            else:
                for prop in model_props.values():
                    prop_path = prop["root_source"]
                    this_path = last_path
                    if prop_path == this_path:
                        current_value[prop["source"]] = get_prop_value(item, prop["pkeys"], prop["source"].split("."))
                if current_path < len(source) - 1:
                    yield from traverse_json(item, current_value, current_path + 1, False)
                else:
                    yield current_value
        elif isinstance(items, list) and is_list_of_dicts(items):
            for item in items:
                yield from traverse_json(item, current_value.copy(), current_path, in_model)

    c_value = {}
    for prop in model_props.values():
        c_value[prop["source"]] = None
    yield from traverse_json(data, current_value=c_value)


def _get_data_xml(url: str, source: str, model_props: dict, namespaces: dict):
    if url.startswith(('http', 'https')):
        f = requests.get(url, timeout=30)
        yield from _parse_xml(io.BytesIO(f.content), source, model_props, namespaces)
    else:
        with pathlib.Path(url).open('rb') as f:
            yield from _parse_xml(f, source, model_props, namespaces)


def _get_data_json(url: str, source: str, model_props: dict):
    if url.startswith(('http', 'https')):
        f = requests.get(url, timeout=30)
        yield from _parse_json(f.text, source, model_props)
    else:
        with pathlib.Path(url).open(encoding='utf-8-sig') as f:
            yield from _parse_json(f.read(), source, model_props)


def _get_prop_full_source(source: str, prop_source: str):
    if prop_source.startswith(".."):
        split = prop_source[1:].count(".")
        split_source = source.split(".")
        value = f'{".".join(split_source[:len(split_source) - split])}'
        if value == '':
            return '.'
        return value
    else:
        return source


def _get_pkeys_if_ref(prop: Property):
    return_list = []
    if isinstance(prop.dtype, Ref):
        for key in prop.dtype.refprops:
            if key.external.name:
                return_list.append(key.external.name)
    return return_list


def _get_dask_dataframe_meta(model: Model):
    dask_meta = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            dask_meta[prop.external.name] = spinta_to_np_dtype(prop.dtype)
    return dask_meta


@commands.getall.register(Context, Model, Json)
def getall(
    context: Context,
    model: Model,
    backend: Json,
    *,
    query: Expr = None,
    resolved_params: ResolvedParams = None,
    **kwargs
) -> Iterator[ObjectData]:
    bases = parametrize_bases(
        context,
        model,
        model.external.resource,
        resolved_params
    )

    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    props = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            root_source = _get_prop_full_source(model.external.name, prop.external.name)
            props[prop.external.name] = {
                "source": prop.external.name,
                "root_source": root_source.split(".") if root_source != "." else ["."],
                "pkeys": _get_pkeys_if_ref(prop)
            }

    meta = _get_dask_dataframe_meta(model)
    df = dask.bag.from_sequence(bases).map(
        _get_data_json,
        source=model.external.name,
        model_props=props
    ).flatten().to_dataframe(meta=meta)
    yield from _dask_get_all(context, query, df, backend, model, builder)


@commands.getall.register(Context, Model, Xml)
def getall(
    context: Context,
    model: Model,
    backend: Xml,
    *,
    query: Expr = None,
    resolved_params: ResolvedParams = None,
    **kwargs
) -> Iterator[ObjectData]:
    bases = parametrize_bases(
        context,
        model,
        model.external.resource,
        resolved_params
    )
    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    props = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            props[prop.name] = {
                "source": prop.external.name,
                "pkeys": _get_pkeys_if_ref(prop)
            }

    meta = _get_dask_dataframe_meta(model)
    df = dask.bag.from_sequence(bases).map(
        _get_data_xml,
        namespaces=_gather_namespaces_from_model(context, model),
        source=model.external.name,
        model_props=props
    ).flatten().to_dataframe(meta=meta)
    yield from _dask_get_all(context, query, df, backend, model, builder)


def _gather_namespaces_from_model(context: Context, model: Model):
    result = {}
    if model.external and model.external.dataset:
        for key, prefix in model.external.dataset.prefixes.items():
            result[key] = prefix.uri
    return result


@commands.getall.register(Context, Model, Csv)
def getall(
    context: Context,
    model: Model,
    backend: Csv,
    *,
    query: Expr = None,
    resolved_params: ResolvedParams = None,
    **kwargs
) -> Iterator[ObjectData]:
    resource_builder = TabularResource(context)
    resource_builder.resolve(model.external.resource.prepare)
    bases = parametrize_bases(
        context,
        model,
        model.external.resource,
        resolved_params,
        model.external.name
    )

    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    df = dask.dataframe.read_csv(list(bases), sep=resource_builder.seperator)
    yield from _dask_get_all(context, query, df, backend, model, builder)


def parametrize_bases(
    context: Context,
    model: Model,
    resource: Resource,
    resolved_params: dict,
    *args
):
    base = resource.external
    if len(resource.source_params) > 0:
        if not resolved_params:
            params = model.external.resource.params
            resolved_params = iterparams(context, model, model.manifest, params)
            for item in resolved_params:
                yield base.format(*args, **item)
        else:
            formatted = base.format(*args, **resolved_params)
            if formatted != base:
                yield formatted
    else:
        yield base


def _dask_get_all(context: Context, query: Expr, df: dask.dataframe, backend: DaskBackend, model: Model, env: DaskDataFrameQueryBuilder):
    keymap: KeyMap = context.get(f'keymap.{model.keymap.name}')

    query = merge_formulas(model.external.prepare, query)
    query = merge_formulas(query, get_enum_filters(context, model))
    query = merge_formulas(query, get_ref_filters(context, model))
    env = env.init(backend, df)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    for i, row in qry.iterrows():
        row = row.to_dict()
        res = {
            '_type': model.model_type(),
        }
        for key, sel in env.selected.items():
            val = _get_row_value(context, row, sel)
            if sel.prop:
                if isinstance(sel.prop.dtype, PrimaryKey):
                    val = keymap.encode(sel.prop.model.model_type(), val)
                elif isinstance(sel.prop.dtype, Ref):
                    val = handle_ref_key_assignment(context, keymap, env, val, sel.prop.dtype)
            res[key] = val
        res = commands.cast_backend_to_python(context, model, backend, res)
        yield res


@commands.wait.register(Context, DaskBackend)
def wait(context: Context, backend: DaskBackend, *, fail: bool = False) -> bool:
    return True


@commands.spinta_to_np_dtype.register(Boolean)
def spinta_to_np_dtype(dtype: Boolean):
    return "boolean"


@commands.spinta_to_np_dtype.register(Number)
def spinta_to_np_dtype(dtype: Number):
    return np.dtype("f")


@commands.spinta_to_np_dtype.register(Integer)
def spinta_to_np_dtype(dtype: Integer):
    return "Int64"


@commands.spinta_to_np_dtype.register(DateTime)
def spinta_to_np_dtype(dtype: DateTime):
    return "object"


@commands.spinta_to_np_dtype.register(DataType)
def spinta_to_np_dtype(dtype: DataType):
    return np.dtype("S")



from typing import Union, Any

from spinta.backends.constants import BackendFeatures
from spinta.components import Property, Page, UrlParams
from spinta.core.ufuncs import Expr, asttoexpr
from spinta.datasets.components import ExternalBackend
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, QueryParams, QueryPage, LiteralProperty
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.types import is_value_literal


def expandable_not_expanded(env: BaseQueryBuilder, prop: Property):
    # If backend does not support expand, assume it is always expanded
    if not env.backend.supports(BackendFeatures.EXPAND):
        return False

    return prop.dtype.expandable and (env.expand is None or (env.expand and prop not in env.expand))


def get_language_column(env: BaseQueryBuilder, dtype: Text):
    default_langs = env.context.get('config').languages
    prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    column = env.backend.get_column(env.table, prop)
    return column


def get_column_with_extra(env: BaseQueryBuilder, prop: Property):
    if isinstance(prop.dtype, Text):
        return get_language_column(env, prop.dtype)
    return env.backend.get_column(env.table, prop)


def get_page_values(env: BaseQueryBuilder, row: dict):
    if not env.page.page_.filter_only:
        if isinstance(env.model.backend, ExternalBackend):
            return [row[item.prop.external.name] for item in env.page.page_.by.values()]
        else:
            return [row[item.prop.name] for item in env.page.page_.by.values()]


def merge_with_page_selected_list(select_list: list, page: QueryPage):
    merged_selected = select_list or []
    if page.page_.enabled and page.select is not None:
        for select in page.select:
            if select not in merged_selected:
                merged_selected.append(select)
    return merged_selected


def merge_with_page_sort(sort: list, page: QueryPage):
    merged_sort = sort or []
    if page.page_.enabled and page.sort:
        merged_sort = page.sort
    return merged_sort


def merge_with_page_limit(limit: int, page: QueryPage):
    if page and page.size and page.page_.enabled:
        if limit and limit < page.size:
            return limit
        return page.size
    return limit


def update_query_with_url_params(query_params: QueryParams, url_params: UrlParams):
    query_params.prioritize_uri = url_params.fmt.prioritize_uri
    query_params.lang_priority = url_params.accept_langs
    query_params.lang = url_params.lang
    query_params.expand = url_params.expand


def _create_or_condition(condition_info: list):
    action = condition_info[0]
    page_by = condition_info[1]
    or_null = condition_info[2]

    result = {
        'name': action,
        'args': [{
            'name': 'bind',
            'args': [page_by.prop.name]
        }, page_by.value]
    }

    if or_null != -1:
        result = {
            'name': 'or',
            'args': [
                result,
                {
                    'name': 'eq' if or_null == 1 else 'ne',
                    'args': [{
                        'name': 'bind',
                        'args': [page_by.prop.name]
                    }, None]
                }
            ]}
    return result


def _get_null_action(by: str, needs_null: bool = False):
    if not needs_null:
        return -1

    if by.startswith('-'):
        return 0
    return 1


def get_pagination_compare_query(
    model_page: Page,
) -> Union[Expr, None]:
    filtered = model_page
    item_count = len(filtered.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    if not filtered.all_none() or not filtered.first_time:
        for i, (by, page_by) in enumerate(filtered.by.items()):
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if page_by.value is not None:
                            if by.startswith('-'):
                                where_list[n].append(('lt', page_by, _get_null_action(by)))
                            else:
                                where_list[n].append(('gt', page_by, _get_null_action(by, True)))
                        else:
                            if by.startswith('-'):
                                where_list[n].append(('ne', page_by, _get_null_action(by)))
                    else:
                        where_list[n].append(('eq', page_by, _get_null_action(by)))
    remove_list = []

    for i, (by, value) in enumerate(filtered.by.items()):
        if value.value is None and not by.startswith('-'):
            remove_list.append(where_list[i])
    for item in remove_list:
        where_list.remove(item)

    where_compare = {}
    for where in where_list:
        compare = {}
        for item in where:
            condition = _create_or_condition(item)
            if compare:
                compare = {
                    'name': 'and',
                    'args': [
                        compare,
                        condition
                    ]
                }
            else:
                compare = condition
        if where_compare:
            where_compare = {
                'name': 'or',
                'args': [
                    where_compare,
                    compare
                ]
            }
        else:
            where_compare = compare

    if where_compare:
        where_compare = asttoexpr(where_compare)

    query = where_compare or None
    return query


def add_page_expr(expr: Expr, page: Page):
    return merge_formulas(expr, asttoexpr({
        'name': 'paginate',
        'args': [
            page
        ]
    }))


def process_literal_value(value: Any) -> Any:
    if is_value_literal(value):
        return LiteralProperty(value)
    return value



from typing import Any

from multipledispatch import dispatch

from spinta.backends.constants import TableType
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.cli.helpers.migrate import MigrateRename
from spinta.utils.itertools import ensure_list


@dispatch(str, str)
def name_changed(
    old_name: str,
    new_name: str
) -> bool:
    return old_name != new_name


@dispatch(str, str, str, str)
def name_changed(
    old_table_name: str,
    new_table_name: str,
    old_property_name: str,
    new_property_name: str
) -> bool:
    return old_table_name != new_table_name or old_property_name != new_property_name


def get_pg_table_name(table_name: str) -> str:
    return get_pg_name(table_name)


def get_pg_changelog_name(table_name: str) -> str:
    return get_pg_name(f"{table_name}{TableType.CHANGELOG.value}")


def get_pg_file_name(table_name: str, arg: str = "") -> str:
    return get_pg_name(f"{table_name}{TableType.FILE.value}{arg}")


def get_pg_column_name(column_name: str) -> str:
    return get_pg_name(column_name)


def get_pg_constraint_name(table_name: str, columns: Any) -> str:
    column_names = ensure_list(columns)
    return get_pg_name(f"{table_name}_{'_'.join(column_names)}_key")


def get_pg_removed_name(name: str) -> str:
    new_name = name.split("/")
    if not new_name[-1].startswith("__"):
        new_name[-1] = f'__{new_name[-1]}'
    new_name = '/'.join(new_name)
    new_name = get_pg_name(new_name)
    return new_name


def get_pg_foreign_key_name(table_name: str, column_name: str) -> str:
    return get_pg_name(f"fk_{table_name}_{column_name.removeprefix('_')}")


def get_pg_index_name(table_name: str, columns: Any) -> str:
    column_names = ensure_list(columns)
    return get_pg_name(f"ix_{table_name}_{'_'.join(column_names)}")


def is_removed(name: str) -> bool:
    if '/' in name:
        last = name.split('/')[-1]
        return last.startswith('__')
    return name.startswith('__')


def nested_column_rename(column_name: str, table_name: str, rename: MigrateRename) -> str:
    renamed = rename.get_column_name(table_name, column_name)
    if name_changed(column_name, renamed):
        return renamed

    if '.' in column_name:
        split_name = column_name.split('.')
        renamed = nested_column_rename('.'.join(split_name[:-1]), table_name, rename)
        return f'{renamed}.{split_name[-1]}'

    return column_name



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import BIGINT, ARRAY

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.constants import BackendFeatures
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import get_root_attr, MigratePostgresMeta
from spinta.backends.postgresql.helpers.migrate.name import name_changed, get_pg_file_name, get_pg_column_name
from spinta.components import Context
from spinta.types.datatype import File
from spinta.utils.schema import NotAvailable


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, File)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: File, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    name = new.prop.name
    pg_name = get_pg_column_name(name)
    nullable = not new.required
    table_name = rename.get_table_name(table.name)
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(get_pg_column_name(f'{name}._id'), sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(get_pg_column_name(f'{name}._content_type'), sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(get_pg_column_name(f'{name}._size'), BIGINT, nullable=nullable)
    ))
    if BackendFeatures.FILE_BLOCKS in new.backend.features:
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(get_pg_column_name(f'{name}._bsize'), sa.Integer, nullable=nullable)
        ))
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(get_pg_column_name(f'{name}._blocks'), ARRAY(pkey_type, ), nullable=nullable)
        ))
    old_table = get_pg_file_name(table.name, f'/{pg_name}')
    new_table = get_pg_file_name(table_name, f'/{pg_name}')
    if not inspector.has_table(old_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, File)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: list, new: File, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    column_name = rename.get_column_name(table.name, new.prop.name)
    old_name = rename.get_old_column_name(table.name, new.prop.name)
    for item in old:
        table_name = rename.get_table_name(table.name)
        nullable = new.required if new.required == item.nullable else None
        item_name = item.name
        new_name = item_name.replace(get_root_attr(item_name), column_name) if not item_name.startswith(
            column_name) else None
        if nullable is not None or new_name is not None:
            handler.add_action(ma.AlterColumnMigrationAction(
                table_name=table_name,
                column_name=item.name,
                nullable=nullable,
                new_column_name=new_name
            ))
    table_name = rename.get_table_name(table.name)
    old_table = get_pg_file_name(table.name, f'/{old_name}')
    new_table_old_prop = get_pg_file_name(table_name,f'/{old_name}')
    new_table_new_prop = get_pg_file_name(table_name, f'/{column_name}')
    if name_changed(old_name, column_name) and inspector.has_table(old_table):
        handler.add_action(
            ma.RenameTableMigrationAction(
                old_table_name=new_table_old_prop,
                new_table_name=new_table_new_prop
            )
        )



import logging
import pathlib
from typing import List, Optional

import requests
from typer import Argument, Option, Typer, echo, Exit
from typer import Context as TyperContext

from spinta import commands, exceptions
from spinta.auth import get_client_id_from_name, get_clients_path
from spinta.cli.helpers.auth import require_auth

from spinta.cli.helpers.errors import ErrorCounter, cli_error
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.cli.helpers.push.utils import extract_dependant_nodes

from spinta.cli.helpers.store import prepare_manifest, attach_backends, attach_keymaps
from spinta.client import get_access_token, get_client_credentials
from spinta.components import Action, Config, Mode
from spinta.core.context import configure_context
from spinta.datasets.keymaps.sync import sync_keymap
from spinta.types.namespace import sort_models_by_ref_and_base

log = logging.getLogger(__name__)

keymap = Typer()


@keymap.command('sync', short_help="Sync keymap from external data source")
def keymap_sync(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    input_source: str = Option(None, '-i', '--input', help=(
        "Input source file"
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Sync only specified dataset"
    )),
    dry_run: bool = Option(False, '--dry-run', help=(
        "Run keymap sync in dry-run mode, no changes will be made."
    )),
    max_error_count: int = Option(50, '--max-errors', help=(
        "If errors exceed given number, keymap command will be stopped."
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    no_progress_bar: bool = Option(False, '--no-progress-bar', help=(
        "Skip counting total rows to improve performance."
    )),
    read_timeout: float = Option(300, '--read-timeout', help=(
        "Timeout for reading a response, default: 5 minutes (300s). The value is in seconds."
    )),
    connect_timeout: float = Option(5, '--connect-timeout', help=(
        "Timeout for connecting, default: 5 seconds."
    )),
):
    """Sync keymap from external data source"""
    if not input_source:
        cli_error(
            "Input source is required."
        )

    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context, full_load=True)
    config: Config = context.get('config')

    if credentials:
        credsfile = pathlib.Path(credentials)
        if not credsfile.exists():
            cli_error(
                f"Credentials file {credsfile} does not exist."
            )
    else:
        credsfile = config.credentials_file

    creds = get_client_credentials(credsfile, input_source)

    manifest = store.manifest
    if dataset and not commands.has_dataset(context, manifest, dataset):
        cli_error(
            str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset))
        )

    ns = commands.get_namespace(context, manifest, '')

    echo(f"Get access token from {creds.server}")
    token = get_access_token(creds)

    client = requests.Session()
    client.headers['Content-Type'] = 'application/json'
    client.headers['Authorization'] = f'Bearer {token}'

    with context:
        auth_client = get_client_id_from_name(get_clients_path(config), config.default_auth_client)
        require_auth(context, auth_client)

        attach_backends(context, store, manifest)
        attach_keymaps(context, store)
        error_counter = ErrorCounter(max_count=max_error_count)

        models = commands.traverse_ns_models(context, ns, manifest, Action.SEARCH, dataset_=dataset, source_check=True)
        models = sort_models_by_ref_and_base(list(models))

        # Synchronize keymaps
        with manifest.keymap as km:
            dependant_models = extract_dependant_nodes(context, models, False)
            sync_keymap(
                context=context,
                keymap=km,
                client=client,
                server=creds.server,
                models=dependant_models,
                error_counter=error_counter,
                no_progress_bar=no_progress_bar,
                reset_cid=True,
                dry_run=dry_run,
                timeout=(connect_timeout, read_timeout)
            )

        if error_counter.has_errors():
            raise Exit(code=1)



from copy import copy
from typing import Any, List
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import TypeVar

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta import commands
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.components import Property, Node
from spinta.datasets.components import Dataset, Resource, ExternalBackend
from spinta.datasets.inspect.helpers import create_manifest_from_inspect
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.utils.schema import NA
from spinta.components import Model


def inspect(
    ctx: TyperContext,
    manifest: Optional[str] = Argument(None, help="Path to manifest."),
    resource: Optional[Tuple[str, str]] = Option(
        (None, None), '-r', '--resource',
        help=(
            "Resource type and source URI "
            "(-r sql sqlite:////tmp/db.sqlite)"
        ),
    ),
    formula: str = Option('', '-f', '--formula', help=(
        "Formula if needed, to prepare resource for reading"
    )),
    backend: Optional[str] = Option(None, '-b', '--backend', help=(
        "Backend connection string"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output tabular manifest in a specified file"
    )),
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
    priority: str = Option('manifest', '-p', '--priority', help=(
        "Merge property priority ('manifest' or 'external')"
    ))
):
    """Update manifest schema from an external data source"""
    if priority not in ['manifest', 'external']:
        echo(f"Priority \'{priority}\' does not exist, there can only be \'manifest\' or \'external\', it will be set to default 'manifest'.")
        priority = 'manifest'

    manifest = convert_str_to_manifest_path(manifest)
    context, manifest = create_manifest_from_inspect(
        context=ctx.obj,
        manifest=manifest,
        resources=resource,
        formula=formula,
        backend=backend,
        auth=auth,
        priority=priority
    )

    if output:
        output_type = output.split(".")[-1]
        config = context.get("config")
        if output_type not in config.exporters:
            echo(f"unknown export file type {output_type!r}")
            raise Exception
        fmt = config.exporters[output_type]
        commands.render(ctx.obj, manifest, fmt, path=output)
    else:
        echo(render_tabular_manifest(context, manifest))



from spinta import commands
from spinta.backends.postgresql.ufuncs.result.components import PgResultBuilder
from spinta.cli.helpers.push.utils import get_data_checksum
from spinta.components import Property
from spinta.core.ufuncs import ufunc, Expr
from spinta.ufuncs.resultbuilder.helpers import get_row_value


@ufunc.resolver(PgResultBuilder, Expr)
def checksum(env: PgResultBuilder, expr: Expr):
    values = {}
    model = None
    if expr.args:
        for arg in expr.args:
            if arg.prop is None or not isinstance(arg.prop, Property):
                continue

            if model is None:
                model = arg.prop.model

            values[arg.prop.place] = commands.cast_backend_to_python(
                env.context,
                arg.prop,
                model.backend,
                get_row_value(env.context, env, env.data, arg)
            )
    return get_data_checksum(values, model)



from typing import Dict, Any
from typing import Iterable
from typing import List

import sqlalchemy as sa
import tqdm

from spinta import commands
from spinta.cli.helpers.push.components import PushRow
from spinta.cli.helpers.push.utils import extract_state_page_keys, construct_where_condition_from_page
from spinta.cli.helpers.push.write import prepare_rows_for_deletion
from spinta.commands.read import PaginationMetaData, get_paginated_values
from spinta.components import Context, pagination_enabled
from spinta.components import Model
from spinta.components import get_page_size


def get_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    no_progress_bar: bool = False,
):
    counts = (
        _get_deleted_row_counts(
            models,
            context,
            metadata
        )
    )
    total_count = sum(counts.values())
    if total_count > 0:
        rows = _iter_deleted_rows(
            models,
            context,
            metadata,
            counts,
            no_progress_bar,
        )
        if not no_progress_bar:
            rows = tqdm.tqdm(rows, 'PUSH DELETED', ascii=True, total=total_count)
        for row in rows:
            yield row


def _get_deleted_row_counts(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
) -> dict:
    counts = {}
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]

        row_count = conn.execute(
            sa.select(sa.func.count(table.c.id)).
            where(
                sa.and_(
                    table.c.pushed.is_(None),
                    table.c.error.is_(False)
                )
            )
        )
        counts[model.name] = row_count.scalar()
    return counts


def _iter_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    no_progress_bar: bool = False,
) -> Iterable[PushRow]:
    models = reversed(models)
    config = context.get('config')
    conn = context.get('push.state.conn')
    for model in models:
        size = get_page_size(config, model)
        table = metadata.tables[model.name]
        total = counts.get(model.name)

        if pagination_enabled(model):
            rows = _get_deleted_rows_with_page(
                context,
                model,
                table,
                size,
            )
        else:
            rows = conn.execute(
                sa.select([table.c.id]).
                where(
                    table.c.pushed.is_(None) &
                    table.c.error.is_(False)
                )
            )
        if not no_progress_bar:
            rows = tqdm.tqdm(
                rows,
                model.name,
                ascii=True,
                total=total,
                leave=False
            )

        for row in rows:
            yield prepare_rows_for_deletion(model, row[table.c.id])


def _get_deleted_rows_with_page(
    context: Context,
    model: Model,
    table: sa.Table,
    size: int,
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    order_by = []
    page = commands.create_page(model.page)
    page.size += 1

    for page_by in page.by.values():
        order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))

    required_where_cond = sa.and_(
        table.c.pushed.is_(None),
        table.c.error.is_(False)
    )

    page_meta = PaginationMetaData(
        page_size=size
    )

    while not page_meta.is_finished:
        page_meta.is_finished = True

        cond = construct_where_condition_from_page(page, table)
        where_cond = None

        if cond is not None:
            where_cond = cond

        if where_cond is not None:
            where_cond = sa.and_(
                required_where_cond,
                where_cond
            )
        else:
            where_cond = required_where_cond

        rows = conn.execute(
            sa.select([table]).
            where(
                where_cond
            ).order_by(*order_by).limit(size)
        )

        yield from get_paginated_values(page, page_meta, rows, extract_state_page_keys)








from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Object


@commands.check.register(Context, Object)
def check(context: Context, dtype: Object):
    for prop in dtype.properties.values():
        commands.check(context, prop)



from typing import Any

from spinta.components import Model, Config, PageInfo
from spinta.core.ufuncs import Env, Expr
from spinta.exceptions import FieldNotInResource
from spinta.ufuncs.loadbuilder.helpers import page_contains_unsupported_keys


class LoadBuilder(Env):
    model: Model

    def resolve(self, expr: Any):
        if not isinstance(expr, Expr):
            # Expression is already resolved, return resolved value.
            return expr

        if expr.name in self._resolvers:
            ufunc = self._resolvers[expr.name]

        else:
            args, kwargs = expr.resolve(self)
            return self.default_resolver(expr, *args, **kwargs)

        if ufunc.autoargs:
            # Resolve arguments automatically.
            args, kwargs = expr.resolve(self)
            try:
                return ufunc(self, *args, **kwargs)
            except NotImplementedError:
                return self.default_resolver(expr, *args, **kwargs)

        else:
            # Resolve arguments manually.
            try:
                return ufunc(self, expr)
            except NotImplementedError:
                pass

    def load_page(self):
        config: Config = self.context.get('config')
        page = PageInfo(
            self.model,
            enabled=config.enable_pagination
        )
        page_given = False

        if self.model.external and self.model.external.prepare:
            resolved = self.resolve(self.model.external.prepare)
            if not isinstance(resolved, list):
                resolved = [resolved]
            for item in resolved:
                if isinstance(item, PageInfo):
                    page = item
                    page_given = True
                    break
        if not page_given:
            args = ['_id']
            if self.model.given.pkeys:
                if isinstance(self.model.given.pkeys, list):
                    args = self.model.given.pkeys
                else:
                    args = [self.model.given.pkeys]
                if '_id' in args:
                    args.remove('_id')
            for arg in args:
                key = arg
                if arg in self.model.properties:
                    prop = self.model.properties[arg]
                    page.keys.update({
                        key: prop
                    })
                else:
                    raise FieldNotInResource(self.model, property=arg)

        # Disable page if given properties are not possible to access
        if page_contains_unsupported_keys(page):
            page.enabled = False

        # Set default page size if nothing was given
        if page.size is None:
            page.size = config.default_page_size

        self.model.page = page





import sqlalchemy as sa

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import MigratePostgresMeta, adjust_kwargs
from spinta.components import Context
from spinta.types.datatype import DataType
from spinta.types.geometry.components import Geometry


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, Geometry)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: DataType, **kwargs):
    columns = commands.prepare(context, backend, new.prop)
    col = None
    for column in columns:
        if isinstance(column, sa.Column):
            col = column
            break
    commands.migrate(context, backend, meta, table, old, col, **kwargs)



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Denorm
from spinta.backends.postgresql.components import PostgreSQL


@commands.prepare.register(Context, PostgreSQL, Denorm)
def prepare(context: Context, backend: PostgreSQL, dtype: Denorm, **kwargs):
    return



from typing import Iterable, AsyncIterator, List, Any

from spinta import commands
from spinta.backends import Backend
from spinta.cli.helpers.data import read_model_data
from spinta.cli.helpers.errors import cli_error
from spinta.cli.helpers.export.components import CounterManager
from spinta.commands.read import get_page
from spinta.components import Context, Model, pagination_enabled, DataItem, Action
from spinta.formats.components import Format
from spinta.ufuncs.basequerybuilder.components import QueryParams


def validate_and_return_shallow_backend(context: Context, type_: str) -> Backend:
    """
    Validates and returns shallow backend.
    Shallow backend, is not loaded and is only used for type checking
    """
    config = context.get("config")
    backends = config.components['backends']
    if type_ not in backends:
        cli_error(
            f"Unknown output backend {type_!r}, from available: {list(backends.keys())}"
        )

    backend = config.components['backends'][type_]()
    backend.type = type_
    return backend


def validate_and_return_formatter(context: Context, type_: str) -> Format:
    config = context.get("config")
    exporters = config.exporters

    if type_ not in exporters:
        cli_error(
            f"Unknown output format {type_!r}, from available: {list(exporters.keys())}"
        )

    fmt = config.exporters[type_]
    return fmt


def get_data(context: Context, model: Model):
    params = QueryParams()
    params.push = True

    if pagination_enabled(model):
        page = commands.create_page(model.page)
        rows = get_page(
            context,
            model,
            model.backend,
            page,
            [],
            params=params
        )
    else:
        rows = read_model_data(
            context,
            model,
            params=params
        )
    yield from rows


def create_data_items(model: Model, backend: Backend, data: Iterable):
    for item in data:
        yield DataItem(
            model=model,
            backend=backend,
            action=Action.INSERT,
            payload=item
        )


async def prepare_data_without_checks(
    context: Context,
    dstream: AsyncIterator[DataItem],
):
    async for data in dstream:
        data.payload = commands.rename_metadata(context, data.payload)
        data.given = commands.load(context, data.model, data.payload)
        yield data


async def export_data(
    context: Context,
    models: List[Model],
    fmt: Any,
    output: str,
    counter: CounterManager,
):
    txn = commands.gen_object_id(context, fmt)
    for model in models:
        data = get_data(context, model)
        await commands.export_data(
            context,
            model,
            fmt,
            data=data,
            path=output,
            counter=counter,
            txn=txn
        )
        counter.close_model(model)



import datetime
import json
import pathlib
from decimal import Decimal

import dateutil.parser
from multipledispatch import dispatch

from spinta.components import Context, Page, Model, Property
from spinta.types.datatype import DataType, DateTime, Number, Date, Time, URL, URI
from spinta.utils.json import fix_data_for_json


@dispatch(Context, Model, dict)
def prepare_data_for_push_state(context: Context, model: Model, data: dict):
    processed_data = {}
    for key, value in data.items():
        processed_data[key] = prepare_data_for_push_state(context, model.flatprops[key], value)
    return processed_data


@dispatch(Context, Property, object)
def prepare_data_for_push_state(context: Context, prop: Property, data: object):
    return prepare_data_for_push_state(context, prop.dtype, data)


@dispatch(Context, DataType, object)
def prepare_data_for_push_state(context: Context, dtype: DataType, data: object):
    return data


@dispatch(Context, DataType, Decimal)
def prepare_data_for_push_state(context: Context, dtype: DataType, data: Decimal):
    return float(data)


@dispatch(Context, DataType, (dict, list))
def prepare_data_for_push_state(context: Context, dtype: DataType, data: dict):
    data = fix_data_for_json(data)
    data = json.dumps(data)
    return data


@dispatch(Context, DataType, pathlib.Path)
def prepare_data_for_push_state(context: Context, dtype: DataType, data: pathlib.Path):
    return str(data)


@dispatch(Context, DateTime, str)
def prepare_data_for_push_state(context: Context, dtype: DateTime, data: str):
    try:
        return datetime.datetime.fromisoformat(data)
    except ValueError:
        return dateutil.parser.parse(data)


@dispatch(Context, Date, str)
def prepare_data_for_push_state(context: Context, dtype: Date, data: str):
    try:
        return datetime.date.fromisoformat(data)
    except ValueError:
        return dateutil.parser.parse(data).date()


@dispatch(Context, Time, str)
def prepare_data_for_push_state(context: Context, dtype: Time, data: str):
    try:
        return datetime.time.fromisoformat(data)
    except ValueError:
        return dateutil.parser.parse(data).time()






from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

from starlette.requests import Request

from spinta import commands
from spinta.auth import AdminToken
from spinta.commands.read import prepare_data_for_response
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.components import Version
from spinta.formats.html.commands import build_template_context
from spinta.formats.html.components import Cell
from spinta.manifests.components import Manifest
from spinta.renderer import render
from spinta.typing import ObjectData


def make_get_request(
    path: str,
    query: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    method: str = 'GET',
) -> Request:
    if not path.startswith('/'):
        path = '/' + path
    req = {
        'type': 'http',
        'method': method,
        'path': path,
        'path_params': {'path': path},
        'headers': [
            (k.lower().encode(), v.encode())
            for k, v in (headers or {}).items()
        ],
    }
    if query:
        req['query_string'] = query.encode()
    return Request(req)


def render_data(
    context: Context,
    manifest: Manifest,
    path: str,
    query: Optional[str],
    data: ObjectData,
    *,
    method: str = 'GET',
    accept: str = 'application/json',
    headers: Optional[Dict[str, str]] = None,
) -> Optional[Dict[str, Any]]:
    context.set('auth.token', AdminToken())

    if headers is None:
        headers = {}
    headers['Accept'] = accept
    request = make_get_request(path, query, headers, method)
    params: UrlParams = commands.prepare(
        context,
        UrlParams(),
        Version(),
        request,
    )
    action = params.action
    model = params.model

    data = next(prepare_data_for_response(context, model, action, params, data, reserved=[
        '_type',
        '_id',
        '_revision'
    ]))

    if params.action in (Action.GETALL, Action.SEARCH):
        data = [data]

    if params.format == 'html':
        resp = _get_html_template_context(context, model, action, params, data)
    else:
        resp = render(context, request, model, params, data, action=action)

    return resp


def _get_html_template_context(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    rows: Iterable[Dict[str, Cell]],
):
    ctx = build_template_context(context, model, action, params, rows)
    resp = next(ctx['data'], None)
    if resp is not None:
        resp = dict(zip(ctx['header'], resp))
    return resp



from collections import defaultdict

import tqdm

from spinta.components import Model


class CounterManager:
    enabled: bool
    total_counter: tqdm.tqdm
    counters: dict[str, dict[str, tqdm.tqdm]]
    totals: dict[str, int]

    def __init__(self, enabled: bool, totals: dict):
        self.enabled = enabled
        self.totals = totals
        self.counters = defaultdict(dict)

        total = sum([value for value in totals.values()])

        self.total_counter = tqdm.tqdm(
            desc="Export total",
            total=total,
            ascii=True,
        ) if enabled else None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enabled:
            self.close_all()

    def add_counter(self, model: Model, desc: str):
        if not self.enabled:
            return

        key = model.model_type()
        self.counters[key][desc] = tqdm.tqdm(
            desc=desc,
            total=self.totals[key],
            ascii=True
        )

    def update_total(self, value: int):
        if not self.enabled:
            return

        self.total_counter.update(value)

    def update_specific(self, model: Model, key: str, value: int):
        if not self.enabled:
            return

        self.counters[model.model_type()][key].update(value)

    def close_all(self):
        if not self.enabled:
            return

        self.total_counter.close()
        for counters in self.counters.values():
            for counter in counters.values():
                counter.close()

    def close_model(self, model: Model):
        if not self.enabled:
            return

        for counter in self.counters[model.model_type()].values():
            counter.close()



import hashlib
import json
from typing import Any
from typing import List
from typing import Union

import msgpack
import sqlalchemy as sa
from spinta import commands

from sqlalchemy.engine.row import Row
from spinta.auth import authorized
from spinta.cli.helpers.push.components import PushRow
from spinta.components import Action, Page, pagination_enabled
from spinta.components import Context
from spinta.components import Model
from spinta.types.datatype import Ref
from spinta.utils.data import take
from spinta.utils.json import fix_data_for_json
from spinta.utils.nestedstruct import flatten, sepgetter


def get_model(row: PushRow) -> Model:
    return row.model


def get_data_checksum(data: dict, model: Model = None):
    data = fix_data_for_json(take(data))
    data = flatten([data], sepgetter(model))
    data = [[k, v] for x in data for k, v in sorted(x.items())]
    data = msgpack.dumps(data, strict_types=True)
    checksum = hashlib.sha1(data).hexdigest()
    return checksum


def extract_state_page_keys(row: Union[dict, Row]):
    result = []

    if isinstance(row, Row):
        row = row._mapping

    for key, value in row.items():
        if 'page.' in key:
            result.append(value)
    return result


def extract_state_page_id_key(row: Union[dict, Row]):
    result = []

    if isinstance(row, Row):
        row = row._mapping

    for key, value in row.items():
        if key == 'id':
            result.append(value)
            break
    return result


def extract_dependant_nodes(context: Context, models: List[Model], filter_pushed: bool):
    extracted_models = [] if filter_pushed else models.copy()
    for model in models:
        if model.base:
            if commands.identifiable(model.base):
                if model.base.parent not in extracted_models:
                    if filter_pushed and not (model.base.parent in models):
                        extracted_models.append(model.base.parent)
                    elif not filter_pushed:
                        extracted_models.append(model.base.parent)

        for prop in model.properties.values():
            if isinstance(prop.dtype, Ref):
                if authorized(context, prop, action=Action.SEARCH):
                    if commands.identifiable(prop):
                        if prop.dtype.model not in extracted_models:
                            if filter_pushed and not (prop.dtype.model in models):
                                extracted_models.append(prop.dtype.model)
                            elif not filter_pushed:
                                extracted_models.append(prop.dtype.model)

    return extracted_models


def load_initial_page_data(
    context: Context,
    metadata: sa.MetaData,
    models: List[Model],
    incremental: bool,
    override_page: dict
) -> dict:
    if not incremental:
        return {}

    conn = context.get('push.state.conn')
    table = metadata.tables['_page']
    result = {}

    for model in models:
        model_name = model.model_type()
        if pagination_enabled(model):
            if model_name in override_page:
                result[model_name] = override_page[model_name]
                continue

            page = conn.execute(
                sa.select([table.c.value]).
                where(
                    table.c.model == model.name
                )
            ).scalar()
            values = json.loads(page) if page else {}
            result[model_name] = values

    return result


def update_model_page_with_new(
    model_page: Page,
    model_table: sa.Table,
    state_row: Any = None,
    data_row: Any = None,
):
    if state_row:
        for by, page_by in model_page.by.items():
            val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
            model_page.update_value(by, page_by.prop, val)
    elif data_row:
        model_page.update_values_from_list(data_row['_page'])


def construct_where_condition_from_page(
    page: Page,
    table: sa.Table,
    prefix: str = 'page.'
):
    item_count = len(page.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    if not page.all_none():
        for i, (by, page_by) in enumerate(page.by.items()):
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if page_by.value is not None:
                            if by.startswith('-'):
                                where_list[n].append(table.c[f"{prefix}{page_by.prop.name}"] < page_by.value)
                            else:
                                where_list[n].append(
                                    sa.or_(
                                        table.c[f"{prefix}{page_by.prop.name}"] > page_by.value,
                                        table.c[f"{prefix}{page_by.prop.name}"] == None
                                    )
                                )
                        else:
                            if by.startswith('-'):
                                where_list[n].append(table.c[f"{prefix}{page_by.prop.name}"] != page_by.value)
                    else:
                        where_list[n].append(table.c[f"{prefix}{page_by.prop.name}"] == page_by.value)

    where_condition = None

    remove_list = []
    for i, (by, value) in enumerate(page.by.items()):
        if value.value is None and not by.startswith('-'):
            remove_list.append(where_list[i])
    for item in remove_list:
        where_list.remove(item)

    for where in where_list:
        condition = None
        for item in where:
            if condition is not None:
                condition = sa.and_(
                    condition,
                    item
                )
            else:
                condition = item
        if where_condition is not None:
            where_condition = sa.or_(
                where_condition,
                condition
            )
        else:
            where_condition = condition

    return where_condition



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends import Backend
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_pg_sequence_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.backends.postgresql.helpers.migrate.migrate import drop_all_indexes_and_constraints, handle_new_file_type, \
    get_prop_names, create_changelog_table, MigratePostgresMeta, \
    MigrateModelMeta, zip_and_migrate_properties
from spinta.backends.postgresql.helpers.migrate.name import name_changed, get_pg_changelog_name, get_pg_file_name, \
    get_pg_column_name, get_pg_constraint_name, get_pg_removed_name, is_removed
from spinta.components import Context, Model
from spinta.types.datatype import File
from spinta.utils.itertools import ensure_list
from spinta.utils.schema import NotAvailable, NA
from sqlalchemy.engine.reflection import Inspector


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, Model)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: sa.Table, new: Model, **kwargs):
    rename = meta.rename
    handler = meta.handler
    inspector = meta.inspector

    columns = list(old.columns)
    new_table_name = rename.get_table_name(old.name)
    old_table_name = old.name

    # Handle table renaming
    _handle_model_rename(
        old_name=old_table_name,
        new_name=new_table_name,
        inspector=inspector,
        handler=handler
    )

    properties = list(new.properties.values())

    model_meta = MigrateModelMeta()
    model_meta.initialize(
        backend=backend,
        table=old,
        columns=columns
    )

    # Handle property migrations
    zip_and_migrate_properties(
        context=context,
        backend=backend,
        old_table=old,
        new_model=new,
        old_columns=columns,
        new_properties=properties,
        meta=meta,
        rename=rename,
        model_meta=model_meta,
        **kwargs
    )

    # Handle model unique constraint
    required_unique_constraints = _handle_model_unique_constraints(
        old_table=old,
        new_model=new,
        table_name=new_table_name,
        inspector=inspector,
        handler=handler
    )

    # Handle property unique constraints
    required_unique_constraints.extend(_handle_property_unique_constraints(
        context=context,
        backend=backend,
        new_model=new
    ))

    # Clean up old multi constraints for non required models (happens when model.ref is changed)
    # Solo unique constraints are handled with property migrate
    _clean_up_old_multi_constraints(
        old_table=old,
        table_name=new_table_name,
        required_unique_constraints=required_unique_constraints,
        inspector=inspector,
        handler=handler
    )

    # Handle JSON migrations, that need to be run at the end
    _handle_json_column_migrations(
        context=context,
        backend=backend,
        meta=meta,
        model_meta=model_meta,
        old_table=old,
        table_name=new_table_name,
        handler=handler,
        **kwargs
    )


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, NotAvailable, Model)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: NotAvailable, new: Model, **kwargs):
    rename = meta.rename
    handler = meta.handler
    inspector = meta.inspector

    table_name = get_pg_name(get_table_name(new))
    pkey_type = commands.get_primary_key_type(context, backend)

    columns = []
    for prop in new.properties.values():
        # Ignore deleted / reserved properties
        if prop.name.startswith('_'):
            continue

        if isinstance(prop.dtype, File):
            columns += handle_new_file_type(context, backend, inspector, prop, pkey_type, handler)
        else:
            cols = commands.prepare(context, backend, prop)
            if isinstance(cols, list):
                for column in cols:
                    if isinstance(column, (sa.Column, sa.Index)):
                        columns.append(column)
            else:
                if isinstance(cols, sa.Column):
                    columns.append(cols)

    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_created', sa.DateTime),
            sa.Column('_updated', sa.DateTime),
            sa.Column("_id", pkey_type, nullable=False, primary_key=True),
            sa.Column("_revision", sa.Text),
            *columns
        ]

    ))

    # Handle new model's unique constraints
    _handle_new_model_unique_constraints(
        new_model=new,
        table_name=table_name,
        handler=handler
    )

    # Create changelog table
    create_changelog_table(context, new, handler, rename)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: sa.Table, new: NotAvailable):
    handler = meta.handler
    inspector = meta.inspector

    old_table_name = old.name

    # Skip already deleted table
    if is_removed(old_table_name):
        return

    removed_table_name = get_pg_removed_name(old_table_name)

    if inspector.has_table(removed_table_name):
        # Drop table if it was already flagged for deletion
        handler.add_action(ma.DropTableMigrationAction(
            table_name=removed_table_name
        ))

        # Drop changelog if it was already flagged for deletion
        removed_changelog_name = get_pg_changelog_name(removed_table_name)
        if inspector.has_table(removed_changelog_name):
            handler.add_action(ma.DropTableMigrationAction(
                table_name=removed_changelog_name
            ))

        # Drop file tables if they were already flagged for deletion
        removed_file_name = get_pg_file_name(removed_table_name)
        for table in inspector.get_table_names():
            if table.startswith(removed_file_name):
                handler.add_action(ma.DropTableMigrationAction(
                    table_name=table
                ))

    handler.add_action(ma.RenameTableMigrationAction(
        old_table_name=old_table_name,
        new_table_name=removed_table_name
    ))
    drop_all_indexes_and_constraints(inspector, old_table_name, removed_table_name, handler)

    # Flag changelog for deletion
    old_changelog_name = get_pg_changelog_name(old_table_name)
    new_changelog_name = get_pg_changelog_name(removed_table_name)
    if inspector.has_table(old_changelog_name):
        handler.add_action(ma.RenameTableMigrationAction(
            old_table_name=old_changelog_name,
            new_table_name=new_changelog_name
        ))
        handler.add_action(ma.RenameSequenceMigrationAction(
            old_name=get_pg_sequence_name(old_changelog_name),
            new_name=get_pg_sequence_name(new_changelog_name)
        ))
        drop_all_indexes_and_constraints(inspector, old_changelog_name, new_changelog_name, handler)

    # Flag file tables for deletion
    for table in inspector.get_table_names():
        old_file_name = get_pg_file_name(old_table_name)
        if table.startswith(old_file_name):
            split = table.split(TableType.FILE.value)
            removed_file_name = get_pg_file_name(removed_table_name, split[1])
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=table,
                new_table_name=removed_file_name
            ))
            drop_all_indexes_and_constraints(inspector, table, removed_file_name, handler)


def _handle_model_rename(
    old_name: str,
    new_name: str,
    inspector: Inspector,
    handler: MigrationHandler
):
    # Do not rename, if name has not been changed
    if not name_changed(old_name, new_name):
        return

    handler.add_action(ma.RenameTableMigrationAction(
        old_table_name=old_name,
        new_table_name=new_name
    ))

    # Handle Changelog table rename
    old_changelog_name = get_pg_changelog_name(old_name)
    if inspector.has_table(old_changelog_name):
        new_changelog_name = get_pg_changelog_name(new_name)
        handler.add_action(ma.RenameTableMigrationAction(
            old_table_name=old_changelog_name,
            new_table_name=new_changelog_name
        )).add_action(ma.RenameSequenceMigrationAction(
            old_name=get_pg_sequence_name(old_changelog_name),
            new_name=get_pg_sequence_name(new_changelog_name)
        ))

    # Handle File table renames
    old_file_name = get_pg_file_name(old_name)
    for table in inspector.get_table_names():
        if table.startswith(old_file_name):
            split = table.split(TableType.FILE.value)
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=table,
                new_table_name=get_pg_file_name(new_name, split[1])
            ))


def _handle_model_unique_constraints(
    old_table: sa.Table,
    new_model: Model,
    table_name: str,
    inspector: Inspector,
    handler: MigrationHandler
) -> list:
    if not new_model.unique:
        return []

    required_unique_constraints = []

    for property_combination in new_model.unique:
        column_name_list = []

        for prop in property_combination:
            for name in get_prop_names(prop):
                column_name_list.append(get_pg_column_name(name))

        constraints = inspector.get_unique_constraints(old_table.name)
        constraint_name = get_pg_constraint_name(table_name, column_name_list)
        required_unique_constraints.append(column_name_list)

        # Skip unique properties, since they are already handled in other parts
        if len(property_combination) == 1 and property_combination[0].dtype.unique:
            continue

        if not any(constraint['name'] == constraint_name for constraint in constraints):
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                table_name=table_name,
                constraint_name=constraint_name,
                columns=column_name_list
            ))
    return required_unique_constraints


def _handle_property_unique_constraints(
    context: Context,
    backend: PostgreSQL,
    new_model: Model
) -> list:
    required_unique_constraints = []
    for prop in new_model.flatprops.values():
        if not prop.dtype.unique:
            continue

        columns = commands.prepare(context, backend, prop.dtype)
        columns = ensure_list(columns)
        column_name_list = []
        for column in columns:
            if isinstance(column, sa.Column):
                column_name_list.append(column.name)

        if column_name_list:
            required_unique_constraints.append(column_name_list)
    return required_unique_constraints


def _clean_up_old_multi_constraints(
    old_table: sa.Table,
    table_name: str,
    required_unique_constraints: list,
    inspector: Inspector,
    handler: MigrationHandler
):
    # Ignore deleted tables
    if old_table.name.startswith('_'):
        return

    unique_constraints = inspector.get_unique_constraints(old_table.name)
    for constraint in unique_constraints:
        if constraint["column_names"] not in required_unique_constraints:
            if not handler.has_constraint_been_dropped(constraint["name"]):
                handler.add_action(ma.DropConstraintMigrationAction(
                    table_name=table_name,
                    constraint_name=constraint["name"]
                ))


def _handle_json_column_migrations(
    context: Context,
    backend: Backend,
    meta: MigratePostgresMeta,
    model_meta: MigrateModelMeta,
    old_table: sa.Table,
    table_name: str,
    handler: MigrationHandler,
    **kwargs
):
    for json_meta in model_meta.json_columns.values():
        if json_meta.new_keys and json_meta.cast_to is None:
            removed_keys = [key for key, new_key in json_meta.new_keys.items() if new_key == get_pg_removed_name(key)]

            if removed_keys == json_meta.keys or json_meta.full_remove:
                commands.migrate(context, backend, meta, old_table, json_meta.column, NA)
            else:
                for old_key, new_key in json_meta.new_keys.items():
                    if new_key in json_meta.keys:
                        handler.add_action(ma.RemoveJSONAttributeMigrationAction(
                            table_name,
                            json_meta.column,
                            new_key
                        ))
                    handler.add_action(ma.RenameJSONAttributeMigrationAction(
                        table_name,
                        json_meta.column,
                        old_key,
                        new_key
                    ))
        elif isinstance(json_meta.cast_to, tuple) and len(json_meta.cast_to) == 2:
            new_column = json_meta.cast_to[0]
            key = json_meta.cast_to[1]
            commands.migrate(context, backend, meta, old_table, json_meta.column, NA)
            commands.migrate(context, backend, meta, old_table, NA, new_column, foreign_key=False,
                             model_meta=model_meta, **kwargs)
            renamed = json_meta.column._copy()
            renamed.name = get_pg_removed_name(json_meta.column.name)
            renamed.key = get_pg_removed_name(json_meta.column.name)
            handler.add_action(
                ma.TransferJSONDataMigrationAction(
                    table_name,
                    renamed,
                    [(key, new_column)]
                )
            )
        # Rename column
        if json_meta.new_name:
            handler.add_action(ma.AlterColumnMigrationAction(
                table_name,
                json_meta.column.name,
                new_column_name=json_meta.new_name
            ))


def _handle_new_model_unique_constraints(
    new_model: Model,
    table_name: str,
    handler: MigrationHandler
):
    if not new_model.unique:
        return

    for property_combination in new_model.unique:
        column_name_list = []
        for prop in property_combination:
            for name in get_prop_names(prop):
                column_name_list.append(get_pg_name(name))
        constraint_name = get_pg_constraint_name(table_name, column_name_list)
        handler.add_action(ma.CreateUniqueConstraintMigrationAction(
            table_name=table_name,
            constraint_name=constraint_name,
            columns=column_name_list
        ))



from spinta.components import UrlParams
from spinta.core.ufuncs import Env


class RequestParamsBuilder(Env):
    params: UrlParams

    def init(self, params: UrlParams):
        return self(
            params=params,
        )






from typing import List, Union

from spinta.core.config import RawConfig, Path, parse_manifest_path, check_if_manifest_valid
from spinta.exceptions import UnknownManifestType
from spinta.manifests.components import ManifestPath


def convert_str_to_manifest_path(manifests: Union[List[str], str]) -> Union[ManifestPath, List[ManifestPath]]:
    if not manifests:
        return []

    new_manifest = []
    rc = RawConfig()
    rc.read([Path('spinta', 'spinta.config:CONFIG')])
    if isinstance(manifests, list):
        for manifest in manifests:
            new_manifest.append(_parse_cli_manifest_path(rc, manifest))
        return new_manifest
    return _parse_cli_manifest_path(rc, manifests)


def _parse_cli_manifest_path(
    rc: RawConfig,
    path: Union[str, ManifestPath],
) -> ManifestPath:
    from spinta.manifests.components import ManifestPath
    if isinstance(path, ManifestPath):
        return path
    if ":" in path:
        split = path.split(":")
        if len(split) >= 2 and not split[1].startswith("/") and not split[1].startswith("\\"):
            if check_if_manifest_valid(rc, split[0]):
                return ManifestPath(type=split[0], path=":".join(split[1:]))
            else:
                raise UnknownManifestType(type=split[0])
    return parse_manifest_path(rc, path)



from __future__ import annotations

import dataclasses
from typing import List, Union, Any, Tuple

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import array_agg
from sqlalchemy.sql.functions import Function

from spinta.backends import get_property_base_model
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.components import Model, Property
from spinta.core.ufuncs import Expr
from spinta.datasets.backends.sql.ufuncs.components import Selected
from spinta.exceptions import PropertyNotFound
from spinta.exceptions import UnknownMethod
from spinta.types.datatype import DataType, Denorm
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, QueryPage, QueryParams, Func
from spinta.ufuncs.basequerybuilder.helpers import merge_with_page_selected_list, merge_with_page_sort, \
    merge_with_page_limit
from spinta.ufuncs.components import ForeignProperty


class PgQueryBuilder(BaseQueryBuilder):
    backend: PostgreSQL
    model: Model
    table: sa.Table
    joins: dict
    columns: List[sa.Column]

    def init(self, backend: PostgreSQL, table: sa.Table, params: QueryParams = None):
        result = self(
            backend=backend,
            table=table,
            selected=None,
            resolved={},
            columns=[],
            joins={},
            from_=table,
            sort=[],
            limit=None,
            offset=None,
            aggregate=False,
            page=QueryPage(),
            group_by=[],
        )
        result.init_query_params(params)
        return result

    def build(self, where):
        if self.selected is None:
            self.call('select', Expr('select'))

        if not self.aggregate:
            self.selected['_id'] = Selected(
                item=self.add_column(self.table.c['_id']),
                prop=self.model.properties['_id']
            )
            self.selected['_revision'] = Selected(
                item=self.add_column(self.table.c['_revision']),
                prop=self.model.properties['_revision']
            )
        merged_selected = merge_with_page_selected_list(self.columns, self.page)
        merged_sorted = merge_with_page_sort(self.sort, self.page)
        merged_limit = merge_with_page_limit(self.limit, self.page)

        qry = sa.select(merged_selected)
        qry = qry.select_from(self.from_)

        if where is not None:
            qry = qry.where(where)

        if merged_sorted:
            qry = qry.order_by(*merged_sorted)

        if merged_limit is not None:
            qry = qry.limit(merged_limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        if self.group_by:
            qry = qry.group_by(*self.group_by)
        return qry

    def add_column(self, column: Union[sa.Column, Function, sa.sql.expression.ColumnElement]) -> int:
        """Returns position in select column list, which is stored in
        Selected.item.
        """
        assert isinstance(column, (sa.Column, Function, sa.sql.expression.ColumnElement)), column
        if column not in self.columns:
            self.columns.append(column)
        return self.columns.index(column)

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(expr=repr(expr(*args, **kwargs)), name=expr.name)

    def get_joined_table(self, prop: ForeignProperty) -> sa.Table:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            lmodel = fpr.left.prop.model
            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(lmodel)
            lrkey = self.backend.get_column(ltable, fpr.left.prop)

            rmodel = fpr.right.prop.model if not isinstance(fpr.right, Denorm) else fpr.right.rel_prop.model
            if rmodel.base:
                rtable = self.get_joined_base_table(rmodel, prop.right.prop.basename, from_Fk=True)
            else:
                rtable = self.backend.get_table(rmodel).alias()

            rpkey = self.backend.get_column(rtable, rmodel.properties['_id'])

            if isinstance(lrkey, list) and not isinstance(rpkey, list):
                if len(lrkey) == 1:
                    lrkey = lrkey[0]

            condition = lrkey == rpkey
            self.joins[fpr.name] = rtable
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[fpr.name]

    def get_joined_table_from_ref(self, prop: ForeignProperty) -> sa.Table:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(fpr.left.prop.model)
            rmodel = fpr.right.prop.model
            rtable = self.backend.get_table(rmodel).alias()
            rpkey = self.backend.get_refprop_columns(rtable, fpr.left.prop, rmodel)
            lrkey = self.backend.get_column(ltable, fpr.left.prop)
            if type(lrkey) != type(rpkey):
                raise Exception("COUNT DONT MATCH")

            if not isinstance(lrkey, list):
                lrkey = [lrkey]
            if not isinstance(rpkey, list):
                rpkey = [rpkey]

            condition = None
            for left, right in zip(lrkey, rpkey):
                if condition is None:
                    condition = left == right
                else:
                    condition = sa.and_(condition, left == right)
            self.joins[fpr.name] = rtable
            if condition is not None:
                self.from_ = self.from_.outerjoin(rtable, condition)
        return self.joins[fpr.name]

    def get_backref_joined_table(self, prop: ForeignProperty, selector: sa.sql.expression) -> sa.sql.expression:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            l_prop = fpr.left.prop
            r_prop = fpr.right.prop
            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(l_prop.model)
            lrkey = self.backend.get_column(ltable, l_prop)
            rpkey = self.backend.get_column(selector, r_prop, override_table=False)
            condition = None
            if isinstance(lrkey, list) and isinstance(rpkey, list):
                if len(lrkey) == len(rpkey):
                    for i, key in enumerate(lrkey):
                        if condition is None:
                            condition = key == rpkey[i]
                        else:
                            condition = sa.and_(condition, key == rpkey[i])
                else:
                    raise Exception("DOESNT MATCH")
            elif type(lrkey) != type(rpkey):
                raise Exception("TYPE DOESNT MATCH")
            else:
                condition = lrkey == rpkey
            self.joins[fpr.name] = selector
            self.from_ = self.from_.outerjoin(selector, condition)
        return self.joins[fpr.name]

    def get_joined_base_table(self, model: Model, prop: str, from_Fk=False):
        inherit_model = model
        base_model = get_property_base_model(inherit_model, prop)

        if not base_model:
            raise PropertyNotFound(model, property=prop)

        ltable = self.backend.get_table(inherit_model)
        lrkey = self.backend.get_column(ltable, inherit_model.properties['_id'])

        rtable = self.backend.get_table(base_model).alias()
        rpkey = self.backend.get_column(rtable, base_model.properties['_id'])

        if base_model.name in self.joins:
            if self.joins[base_model.name] == rtable:
                return self.joins[base_model.name]

        condition = lrkey == rpkey
        self.joins[base_model.name] = rtable
        if not from_Fk:
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[base_model.name]

    def generate_backref_select(self, left: Property, right: Property, required_columns: List[Tuple[str, str]],
                                aggregate: bool = False):
        select_columns = []
        group_by = []

        left_table = self.backend.get_table(left.model)
        left_keys = self.backend.get_column(left_table, left)

        join_condition = None
        right_table = self.backend.get_table(right)
        right_list_table = None
        list_join_condition = None
        right_keys = self.backend.get_column(right_table, right, override_table=False)
        right_list_keys = None
        if right.list is not None:
            right_list_table = self.backend.get_table(right, TableType.LIST)
            list_join_condition = right_list_table.c['_rid'] == right_table.c['_id']
            right_list_keys = self.backend.get_column(right_list_table, right)
        from_ = right_list_table if right_list_table is not None else right_table
        initial_join_right_keys = right_list_keys if right_list_keys is not None else right_keys

        if aggregate:
            json_columns = []
            for col, label in required_columns:
                json_columns.extend([col, right_table.c[col]])
            column = array_agg(sa.func.json_build_object(*json_columns))
            column = column.label(left.name)
            select_columns.append(column)

            if isinstance(initial_join_right_keys, list):
                for key in initial_join_right_keys:
                    group_by.append(key)
            else:
                group_by.append(initial_join_right_keys)

        else:
            for col, label in required_columns:
                column = right_table.c[col]
                column = column.label(label)
                select_columns.append(column)

        if isinstance(initial_join_right_keys, list):
            select_columns.extend(initial_join_right_keys)
        else:
            select_columns.append(initial_join_right_keys)

        if isinstance(left_keys, list) and isinstance(initial_join_right_keys, list):
            if len(left_keys) == len(initial_join_right_keys):
                for i, key in enumerate(left_keys):
                    if join_condition is None:
                        join_condition = key == initial_join_right_keys[i]
                    else:
                        join_condition = sa.and_(join_condition, key == initial_join_right_keys[i])
            else:
                raise Exception("DOESNT MATCH")
        elif type(left_keys) != type(initial_join_right_keys):
            raise Exception("TYPE DOESNT MATCH")
        else:
            join_condition = left_keys == initial_join_right_keys

        from_ = from_.outerjoin(left_table, join_condition)
        if list_join_condition is not None:
            from_ = from_.outerjoin(right_table, list_join_condition)
        stmt = sa.select(select_columns)
        stmt = stmt.select_from(from_)
        stmt = stmt.group_by(*group_by)
        stmt = stmt.subquery()
        return stmt


@dataclasses.dataclass
class Lower(Func):
    dtype: DataType = None


@dataclasses.dataclass
class Negative(Func):
    arg: Any


@dataclasses.dataclass
class Positive(Func):
    arg: Any


@dataclasses.dataclass
class Recurse(Func):
    args: List[Union[DataType, Func]] = None


# This might be a hack, to access `_base` property
# should probably be remade to work with ForeignProperty
class InheritForeignProperty:

    def __init__(
        self,
        model: Model,
        prop_name: str,
        base_prop: Property
    ):
        self.model = model
        self.base_prop = base_prop
        self.prop_name = prop_name



from spinta import commands
from spinta.components import Context
from spinta.types.text.components import Text


@commands.check.register(Context, Text)
def check(context: Context, dtype: Text):
    for lang in dtype.langs.values():
        commands.check(context, lang)



from typing import List

import sqlalchemy as sa

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.backends.postgresql.helpers.migrate.migrate import name_key, MigratePostgresMeta, adjust_kwargs, \
    is_name_complex, extract_literal_name_from_column, generate_type_missmatch_exception_details, \
    extract_sqlalchemy_columns, is_internal, split_columns, get_spinta_primary_keys, remap_and_rename_columns, \
    remove_property_prefix_from_column_name, zip_and_migrate_properties
from spinta.backends.postgresql.helpers.migrate.name import get_pg_column_name
from spinta.cli.helpers.migrate import MigrateRename
from spinta.components import Context
from spinta.datasets.inspect.helpers import zipitems
from spinta.exceptions import MigrateScalarToRefTooManyKeys, MigrateScalarToRefTypeMissmatch
from spinta.types.datatype import Ref, ExternalRef
from spinta.utils.itertools import ensure_list
from spinta.utils.schema import NotAvailable, NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: Ref, **kwargs):
    columns = commands.prepare(context, backend, new.prop)
    if not isinstance(columns, list):
        columns = [columns]
    for column in columns:
        if isinstance(column, sa.Column):
            commands.migrate(context, backend, meta, table, old, column, **adjust_kwargs(kwargs, 'foreign_key', True))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: Ref, **kwargs):
    commands.migrate(context, backend, meta, table, [old], new, **kwargs)


def _migrate_scalar_to_ref_4(
    context: Context,
    backend: PostgreSQL,
    table: sa.Table,
    columns: List[sa.Column],
    ref: Ref,
    ref_column: sa.Column,
    meta: MigratePostgresMeta,
    handler: MigrationHandler,
    rename: MigrateRename,
    **kwargs
) -> bool:
    """Checks and converts scalar to internal ref

    Args:
        table: old table
        columns: list of old columns
        ref: new Ref property
        ref_column: new Ref converted to column

    Returns:
        bool: Applied scalar to ref conversion or not

    In order to do scalar to ref conversion, there are strict requirements
    otherwise it would be nearly impossible to guess primary key and foreign key mapping
        1. refprops size has to be 1 (since you try to map single column, cannot map to more than that)
        2. column type has to match ref tables primary key type
    This triggers only if column name (after rename) matches ref property name_key
    """

    # Check if there is only 1 primary key to match
    if not len(columns) == 1:
        return False

    table_name = rename.get_table_name(table.name)

    column = columns[0]
    new_name = rename.get_column_name(table.name, column.name)

    # Check if after rename column becomes ref itself, or only part of it (can check if name contains special characters)
    if is_name_complex(new_name):
        return False

    # Check if refprops is size of 1
    if len(ref.refprops) > 1:
        raise MigrateScalarToRefTooManyKeys(ref, primary_keys=[pkey.name for pkey in ref.refprops])

    target = ref.refprops[0]
    target_column = commands.prepare(context, backend, target)
    key = target.name
    old_type = extract_literal_name_from_column(column)
    new_type = extract_literal_name_from_column(target_column)

    # Check if types match
    if old_type != new_type:
        raise MigrateScalarToRefTypeMissmatch(
            ref,
            details=generate_type_missmatch_exception_details(
                [
                    (
                        (column.name, old_type),
                        (key, new_type)
                    )
                ]
            ))
    # Create new empty ref column
    commands.migrate(context, backend, meta, table, NA, ref_column, **kwargs)

    # Apply conversion from scalar to ref column
    handler.add_action(ma.UpgradeTransferDataMigrationAction(
        table_name=table_name,
        referenced_table_name=get_table_name(ref.model),
        ref_column=ref_column,
        columns={
            key: column
        }
    ), True)

    # Drop old column after migration
    commands.migrate(context, backend, meta, table, column, NA, **kwargs)
    return True


def _migrate_scalar_to_ref_3(
    context: Context,
    backend: PostgreSQL,
    table: sa.Table,
    columns: List[sa.Column],
    ref: ExternalRef,
    ref_columns: List[sa.Column],
    meta: MigratePostgresMeta,
    handler: MigrationHandler,
    rename: MigrateRename,
    **kwargs
) -> bool:
    """Checks and converts scalar to external ref

    Args:
        table: old table
        columns: list of old columns
        ref: new ExternalRef property
        ref_columns: new ExternalRef converted to column

    Returns:
        bool: Applied scalar to ref conversion or not

    In order to do scalar to ref conversion, there are strict requirements
    otherwise it would be nearly impossible to guess primary key and foreign key mapping
        1. refprops size has to be 1 (since you try to map single column, cannot map to more than that)
        2. column type has to match ref tables primary key type
    This triggers only if column name (after rename) matches ref property name_key
    """

    # Check if there is only 1 primary key to match
    if not len(columns) == 1:
        return False

    table_name = rename.get_table_name(table.name)

    column = columns[0]
    new_name = rename.get_column_name(table.name, column.name)

    # Check if after rename column becomes ref itself, or only part of it (can check if name contains special characters)
    if is_name_complex(new_name):
        return False

    # Check if refprops is size of 1
    if len(ref.refprops) > 1 or len(ref_columns) > 1:
        raise MigrateScalarToRefTooManyKeys(ref, primary_keys=[pkey.name for pkey in ref.refprops])

    ref_column = ref_columns[0]
    old_type = extract_literal_name_from_column(column)
    new_type = extract_literal_name_from_column(ref_column)

    # Check if types match
    if old_type != new_type:
        raise MigrateScalarToRefTypeMissmatch(
            ref,
            details=generate_type_missmatch_exception_details(
                [
                    (
                        (column.name, old_type),
                        (ref_column.name, new_type)
                    )
                ]
            ))
    # Create new empty ref column
    commands.migrate(context, backend, meta, table, NA, ref_column, **kwargs)

    # Apply conversion from scalar to ref column
    target = remove_property_prefix_from_column_name(
        ref_column.name,
        ref.prop
    )
    handler.add_action(
        ma.DowngradeTransferDataMigrationAction(
            table_name=table_name,
            referenced_table_name=get_table_name(ref.model),
            source_column=column,
            columns={
                ref_column.name: sa.Column(target, type_=ref_column.type)
            },
            target=target
        ), True
    )

    # Drop old column after migration
    commands.migrate(context, backend, meta, table, column, NA, **kwargs)
    return True


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: Ref, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler
    adjusted_kwargs = adjust_kwargs(kwargs, 'foreign_key', True)

    new_primary_columns = commands.prepare(context, backend, new.prop, propagate=False)
    new_primary_columns = ensure_list(new_primary_columns)
    new_primary_columns = extract_sqlalchemy_columns(new_primary_columns)
    new_primary_column_names = [column.name for column in new_primary_columns]

    # Since its `Ref` type, it should only generate 1 columns 'column._id'
    primary_column = new_primary_columns[0]

    new_all_columns = commands.prepare(context, backend, new.prop)
    new_all_columns = ensure_list(new_all_columns)
    new_all_columns = extract_sqlalchemy_columns(new_all_columns)

    new_children_columns = [column for column in new_all_columns if column.name not in new_primary_column_names]
    new_children_column_names = [column.name for column in new_children_columns]

    table_name = rename.get_table_name(table.name)
    old_ref_table = rename.get_old_table_name(get_table_name(new.model))
    old_prop_name = rename.get_old_column_name(table.name, get_column_name(new.prop))

    new_name = get_pg_column_name(new.prop.place)
    ref_model_primary_keys = get_spinta_primary_keys(
        table_name=old_ref_table,
        model=new.model,
        inspector=inspector
    )
    old_columns_internal = is_internal(old, old_prop_name)
    old_primary_columns, old_children_columns = split_columns(
        columns=old,
        base_name=old_prop_name,
        target_base_name=new_name,
        ref_table_primary_key_names=ref_model_primary_keys,
        target_primary_column_names=new_primary_column_names,
        target_children_column_names=new_children_column_names,
        internal=old_columns_internal
    )

    if old_columns_internal:
        # Handle internal ref migration
        commands.migrate(
            context,
            backend,
            meta,
            table,
            old_primary_columns[0],
            primary_column,
            **adjusted_kwargs
        )
    else:
        # Try to do scalar to ref migration
        migrated = _migrate_scalar_to_ref_4(
            context=context,
            backend=backend,
            table=table,
            columns=old_primary_columns,
            ref=new,
            ref_column=primary_column,
            meta=meta,
            handler=handler,
            rename=rename,
            **adjusted_kwargs
        )

        # Scalar did not pass
        if not migrated:
            column_mapping = dict()
            for column in old_primary_columns:
                new_name = rename.get_column_name(table.name, column.name)
                key = new_name.split('.')[-1]
                column_mapping[key] = column

            # Create empty ref column
            commands.migrate(context, backend, meta, table, NA, primary_column, **adjusted_kwargs)

            # Migrate from level 3 to level 4 ref
            handler.add_action(ma.UpgradeTransferDataMigrationAction(
                table_name=table_name,
                referenced_table_name=get_table_name(new.model),
                ref_column=primary_column,
                columns=column_mapping
            ), True)

            # Drop old columns
            for column in column_mapping.values():
                commands.migrate(context, backend, meta, table, column, NA, **adjusted_kwargs)

    zip_and_migrate_properties(
        context=context,
        backend=backend,
        old_table=table,
        new_model=new.prop.model,
        old_columns=old_children_columns,
        new_properties=list(new.properties.values()),
        meta=meta,
        rename=rename,
        root_name=new.prop.place,
        **kwargs
    )


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, ExternalRef)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: ExternalRef, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    adjusted_kwargs = adjust_kwargs(kwargs, 'foreign_key', True)

    table_name = rename.get_table_name(table.name)
    old_ref_table = rename.get_old_table_name(get_table_name(new.model))
    old_prop_name = rename.get_old_column_name(table.name, get_column_name(new.prop))

    new_primary_columns = commands.prepare(context, backend, new.prop, propagate=False)
    new_primary_columns = ensure_list(new_primary_columns)
    new_primary_columns = extract_sqlalchemy_columns(new_primary_columns)
    new_primary_column_name_mapping = {column.name: column for column in new_primary_columns}

    new_all_columns = commands.prepare(context, backend, new.prop)
    new_all_columns = ensure_list(new_all_columns)
    new_all_columns = extract_sqlalchemy_columns(new_all_columns)

    new_children_columns = [column for column in new_all_columns if
                               column.name not in new_primary_column_name_mapping]
    new_children_column_names = [column.name for column in new_children_columns]

    new_name = get_pg_column_name(new.prop.place)
    ref_model_primary_keys = get_spinta_primary_keys(
        table_name=old_ref_table,
        model=new.model,
        inspector=inspector
    )
    old_columns_internal = is_internal(old, old_prop_name)
    old_primary_columns, old_children_columns = split_columns(
        columns=old,
        base_name=old_prop_name,
        target_base_name=new_name,
        ref_table_primary_key_names=ref_model_primary_keys,
        target_primary_column_names=list(new_primary_column_name_mapping.keys()),
        target_children_column_names=new_children_column_names,
        internal=old_columns_internal
    )

    # Check to see if migration has already been achieved
    migrated = False
    if len(old_primary_columns) == 1:
        old_primary_column = old_primary_columns[0]

        if old_columns_internal:
            # Handle Internal ref mapping
            # Ensure columns exist
            column_mapping = {}
            for column in new_primary_columns:
                column_mapping[column.name] = sa.Column(
                    remove_property_prefix_from_column_name(
                        column.name,
                        new.prop
                    ),
                    type_=column.type
                )
                commands.migrate(context, backend, meta, table, NA, column, **adjusted_kwargs)

            # Downgrade ref column
            handler.add_action(
                ma.DowngradeTransferDataMigrationAction(
                    table_name=table_name,
                    referenced_table_name=get_table_name(new.model),
                    source_column=old_primary_column,
                    columns=column_mapping,
                    target='_id'
                ), True
            )

            # Drop old column
            commands.migrate(context, backend, meta, table, old_primary_column, NA, **adjusted_kwargs)
            migrated = True
        else:
            migrated = _migrate_scalar_to_ref_3(
                context=context,
                backend=backend,
                table=table,
                columns=old_primary_columns,
                ref=new,
                ref_columns=new_primary_columns,
                meta=meta,
                handler=handler,
                rename=rename,
                **adjusted_kwargs
            )

    # If no ref migrations were done, try to zip the primary columns and migrate them
    if not migrated:
        renamed_old_primary_columns = remap_and_rename_columns(
            base_name=old_prop_name,
            columns=old_primary_columns,
            table_name=table.name,
            ref_table_name=old_ref_table,
            rename=rename
        )

        zipped_items = zipitems(
            renamed_old_primary_columns.keys(),
            new_primary_column_name_mapping.keys(),
            name_key
        )
        for zipped_item in zipped_items:
            for old_column, new_column in zipped_item:
                if old_column is not NA:
                    old_column = renamed_old_primary_columns[old_column]
                if new_column is not NA:
                    new_column = new_primary_column_name_mapping[new_column]

                if old_column is not None and new_column is None:
                    commands.migrate(context, backend, meta, table, old_column, NA, **adjusted_kwargs)
                else:
                    commands.migrate(context, backend, meta, table, old_column, new_column, **adjusted_kwargs)

    zip_and_migrate_properties(
        context=context,
        backend=backend,
        old_table=table,
        new_model=new.prop.model,
        old_columns=old_children_columns,
        new_properties=list(new.properties.values()),
        meta=meta,
        rename=rename,
        root_name=new.prop.place,
        **kwargs
    )



import logging

from spinta import commands
from spinta.components import Context, Namespace
from spinta.manifests.helpers import init_manifest, _configure_manifest, load_manifest_nodes
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.manifests.components import Manifest
from spinta.manifests.internal_sql.helpers import read_initial_schema, load_internal_manifest_nodes, read_schema

log = logging.getLogger(__name__)


@commands.create_request_manifest.register(Context, InternalSQLManifest)
def create_request_manifest(context: Context, manifest: InternalSQLManifest):
    old = manifest
    store = manifest.store
    manifest = old.__class__()
    rc = context.get('rc')
    init_manifest(context, manifest, old.name)
    _configure_manifest(
        context, rc, store, manifest,
        backend=store.manifest.backend.name if store.manifest.backend else None,
    )
    commands.load(context, manifest)
    commands.link(context, manifest)
    return manifest


@commands.load_for_request.register(Context, InternalSQLManifest)
def load_for_request(context: Context, manifest: InternalSQLManifest):
    context.attach('transaction.manifest', manifest.transaction)
    schemas = read_initial_schema(context, manifest)
    load_internal_manifest_nodes(context, manifest, schemas, link=True)


@commands.load.register(Context, InternalSQLManifest)
def load(
    context: Context,
    manifest: InternalSQLManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load=False
):
    if load_internal:
        target = into or manifest
        if '_schema' not in target.get_objects()['model']:
            store = context.get('store')
            commands.load(context, store.internal, into=target, full_load=full_load)

    if full_load:
        schemas = read_schema(manifest.path)
        if into:
            log.info(
                'Loading freezed manifest %r into %r from %s.',
                manifest.name,
                into.name,
                manifest.path,
            )
            load_manifest_nodes(context, into, schemas, source=manifest)
        else:
            log.info(
                'Loading freezed manifest %r from %s.',
                manifest.name,
                manifest.path,
            )
            load_manifest_nodes(context, manifest, schemas)

        for source in manifest.sync:
            commands.load(
                context, source,
                into=into or manifest,
                freezed=freezed,
                rename_duplicates=rename_duplicates,
                load_internal=load_internal,
                full_load=full_load
            )




import logging
from typer import Context as TyperContext
from typer import Option

from spinta.cli.helpers.store import load_config
from spinta.cli.helpers.upgrade.scripts import run_specific_script, run_all_scripts, get_all_upgrade_script_names, \
    does_script_exist
from spinta.core.context import configure_context
from spinta.exceptions import UpgradeError, UpgradeScriptNotFound

log = logging.getLogger(__name__)


def upgrade(
    ctx: TyperContext,
    ensure_config_dir: bool = Option(True, '--ensure-config', help=(
        "Ensures that all config files are created."
    )),
    force: bool = Option(False, '-f', '--force', help=(
        "Skips all checks when running upgrades."
    )),
    destructive: bool = Option(False, '-d', '--destructive', help=(
        """
        Runs scripts in destructive mode (scripts can now override already migrated files).
        WARNING: only use this if you know what will be changed (recommended to create backups).
        """
    )),
    run: str = Option(None, '-r', '--run', help=(
        f"""
        Specify a script to run.
        Available scripts: {get_all_upgrade_script_names()}
        """
    )),
):
    context = configure_context(ctx.obj)
    try:
        load_config(
            context,
            ensure_config_dir=ensure_config_dir
        )
    except UpgradeError:
        # Ignore UpgradeErrors, since this functions handles them
        pass

    if run is not None:
        if not does_script_exist(run):
            raise UpgradeScriptNotFound(
                script=run,
                available_scripts=get_all_upgrade_script_names()
            )

        run_specific_script(
            context=context,
            destructive=destructive,
            force=force,
            script_name=run
        )
        return

    run_all_scripts(
        context=context,
        destructive=destructive,
        force=force
    )



from spinta.manifests.components import Manifest


class MermaidManifest(Manifest):
    type = 'mmd'

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.mmd')
